/* Copyright (C) 2005-2006  Emiliano Betti and Marco Cesati
   Copyright (C) 2003-2004  Marco Cesati
   Copyright (C) 2002-2003  Shell Technologies s.r.l.

   Authors:
            Emiliano Betti <emilianobetti@gmail.com>
            Marco Cesati   <marcocesati@gmail.com>

   Driver for the Galil 1800 PCI motion controllers (http://www.galilmc.com/)

   This is a driver for the Linux kernel 2.6. It is based on an older driver
   for the kernel 2.4, which in turn was loosely inspired by Galil's own
   Linux driver, as well as by a bunch of PCI drivers in the kernel source code

   This program is free software; you can redistribute  it and/or modify it
   under  the terms of  the GNU General  Public License as published by the
   Free Software Foundation.
*/

#ifndef _GALIL1800_CARD_H_
#define _GALIL1800_CARD_H_

#ifndef __KERNEL__
#error UserMode programs should include galil1800.h rather than galil1800_card.h
#endif

/* Default size of write buffer (in bytes) */
#define GALIL1800_CARD_WBUFF_DEFAULT_SZ (4096*2)
/* Min. size of write buffer (in bytes) */
#define GALIL1800_CARD_WBUFF_SZ_MIN (GALIL1800_WRITE_FIFO_SIZE*2)
/* Max size of write buffer (in bytes) */
#define GALIL1800_CARD_WBUFF_SZ_MAX (4096*1024)

/* enumeration for means of flag bit
  (last is the number of flag bit used for
   each byte in the buffer: pay attention to
   set only values that are power of 2 and if you
   set to more than 8, you must check other code!!) */
enum galil1800_card_wb_flag_bit { GALIL1800_CARD_WB_FLAG_PRESENT=0,
	GALIL1800_CARD_WB_FLAG_PAUSE, GALIL1800_CARD_WB_FLAG_BIT_FOR_BYTE };

/* Size of flag bitmap for write buffer (in bytes) */
#define GALIL1800_CARD_WB_FLAG_BITMAP_SZ(_wbuff) \
    ((((_wbuff)->size)*GALIL1800_CARD_WB_FLAG_BIT_FOR_BYTE)/8)


/* Pseudo command "pause" (by ioctl) enable (1) or disable(0) */
#define GALIL1800_CARD_PAUSE_ENABLE_DEFAULT 0


/* The circular card device write buffer:
    N.B.: work only whit one writer and one reader!
    - Flag Present == 1  => byte ready to read
    - Flag Pause == 1  => byte for pause timer.
                          Each pause take four byte, so,
                          after the first, also the next
                          3 bytes must have this flag set.
    - For reading: until (tail->flagPresent == 1 && !pause) =>
                                                     read tail, incr. tail
    - For writing: until (head->flagPresent == 0) => write head, incr. head */
struct galil1800_card_wbuffer {
	unsigned short run; /* used to stop the workqueue (1 run, 0 stop) */
	unsigned int size; /* if 0 write buffer disable,
			      otherwise the buffer size in byte */

	unsigned long *flag; /* Map bit for flag for each byte of data */
	unsigned char *data; /* Pointer to first byte of buffer */
	unsigned char *head; /* Pointer to first byte to write */
	unsigned char *tail; /* Pointer to first byte to read */

	/* Next two members are modified and used only for event report.
	See function galil1800_add_wbuff_event() in galil1800_card.c */
	unsigned char last_size_p10; /* last size in per 10 value */
	unsigned char last_event_opt_flag; /* last size in per 10 value */

	/* Next 3 members are used to manage pause command */
	unsigned int pause_enable; /* if 0 pause pseudo-cmd disable, 1 enable */
	unsigned long pause; /* if 0 pause not set,
				otherwise the current pause wake up jiffie */
	struct timer_list pause_timer; /* Timer for manage "pause" pseudo-cmd */

	wait_queue_head_t read_wait;  /* stop reading:
					 waiting for data && !pause */
	wait_queue_head_t write_wait; /* waiting for write fifo is not full */

	struct workqueue_struct *wq; /* workqueue for flush the buffer */
	struct work_struct work;  /* work that flush the buffer */

	/* Class Device Attribute for set "size" and "pause_enable" */
	struct class_device_attribute cd_attr_size;
	struct class_device_attribute cd_attr_pause_enable;
};

#define GALIL1800_CARD_WB_CURRENT_SIZE(_wb_p)                \
	( (_wb_p)->head < (_wb_p)->tail ?                    \
	((_wb_p)->size + (_wb_p)->head - (_wb_p)->tail) :    \
		((_wb_p)->head - (_wb_p)->tail) )


/* First step of initialization:
   must be call in probe function */
#define init_galil1800_card_wbuffer(_wbuff_p, _size, _pause_onoff) do {    \
	(_wbuff_p)->run = 0;                                               \
	(_wbuff_p)->size = (_size);                                        \
	(_wbuff_p)->flag = 0;                                              \
	(_wbuff_p)->data = 0;                                              \
	(_wbuff_p)->head = 0;                                              \
	(_wbuff_p)->tail = 0;                                              \
	(_wbuff_p)->last_size_p10 = 0;                                     \
	(_wbuff_p)->last_event_opt_flag = 0;                               \
	(_wbuff_p)->pause_enable = (_pause_onoff);                         \
	(_wbuff_p)->pause = 0UL;                                           \
	init_timer(&((_wbuff_p)->pause_timer));                            \
	(_wbuff_p)->pause_timer.data=(unsigned long)(_wbuff_p);            \
	(_wbuff_p)->pause_timer.function=galil1800_card_pause_wake_up;     \
	init_waitqueue_head(&(_wbuff_p)->read_wait);                       \
	init_waitqueue_head(&(_wbuff_p)->write_wait);                      \
	(_wbuff_p)->wq = 0;                                                \
	if((_wbuff_p)->pause_enable)                                       \
		INIT_WORK( &((_wbuff_p)->work),                            \
			galil1800_card_wbuffer_flush_whit_pause_cmd,       \
			(_wbuff_p));                                       \
	else                                                               \
		INIT_WORK( &((_wbuff_p)->work),                            \
			galil1800_card_wbuffer_flush, (_wbuff_p));         \
}while(0)


#define GALIL1800_CARD_WB_FLAG_BIT_OFFSET(_wbuff_p, _pt_byte, _flag_type)  \
	( (((_pt_byte) - (_wbuff_p)->data) *                               \
		GALIL1800_CARD_WB_FLAG_BIT_FOR_BYTE) + (_flag_type) )

#define GALIL1800_CARD_WB_FLAG_PRESENT_OFFSET(_wbuff_p, _pt_byte)  \
	GALIL1800_CARD_WB_FLAG_BIT_OFFSET(_wbuff_p, _pt_byte,      \
				GALIL1800_CARD_WB_FLAG_PRESENT)

#define GALIL1800_CARD_WB_FLAG_PAUSE_OFFSET(_wbuff_p, _pt_byte)    \
	GALIL1800_CARD_WB_FLAG_BIT_OFFSET(_wbuff_p, _pt_byte,      \
				GALIL1800_CARD_WB_FLAG_PAUSE)


#define GALIL1800_CARD_WB_FULL(_wbuff_p)                             \
	test_bit(GALIL1800_CARD_WB_FLAG_PRESENT_OFFSET((_wbuff_p),   \
				(_wbuff_p)->head), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_TEST_PRESENT_FLAG(_wbuff_p, _pt)           \
	test_bit(GALIL1800_CARD_WB_FLAG_PRESENT_OFFSET((_wbuff_p),   \
					(_pt)), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_SET_PRESENT_FLAG(_wbuff_p)                 \
	set_bit(GALIL1800_CARD_WB_FLAG_PRESENT_OFFSET((_wbuff_p),    \
				(_wbuff_p)->head), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_SET_PAUSE_FLAG(_wbuff_p)                   \
	set_bit(GALIL1800_CARD_WB_FLAG_PAUSE_OFFSET((_wbuff_p),      \
				(_wbuff_p)->head), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_NOT_EMPTY(_wbuff_p)                        \
	test_bit(GALIL1800_CARD_WB_FLAG_PRESENT_OFFSET((_wbuff_p),   \
				(_wbuff_p)->tail), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_CLEAR_PRESENT_FLAG(_wbuff_p)               \
	clear_bit(GALIL1800_CARD_WB_FLAG_PRESENT_OFFSET((_wbuff_p),  \
				(_wbuff_p)->tail), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_TEST_PAUSE(_wbuff_p)                       \
	test_bit(GALIL1800_CARD_WB_FLAG_PAUSE_OFFSET((_wbuff_p),     \
				(_wbuff_p)->tail), (_wbuff_p)->flag)

#define GALIL1800_CARD_WB_CLEAR_PAUSE_FLAG(_wbuff_p)                 \
	clear_bit(GALIL1800_CARD_WB_FLAG_PAUSE_OFFSET((_wbuff_p),    \
				(_wbuff_p)->tail), (_wbuff_p)->flag)


/* Increment a poiter to data in buffer */
#define galil1800_card_wbuffer_inc_pt(_wbuff_p, _pt) do {              \
	(_pt) = ((_pt) + 1) < ((_wbuff_p)->data + (_wbuff_p)->size) ?  \
		((_pt) + 1) : (_wbuff_p)->data;                        \
}while(0)

/* Add one byte without space check */
#define __galil1800_card_wbuffer_add_byte(_wbuff_p, _byte_info) do {   \
	*((_wbuff_p)->head)=(_byte_info);                              \
	galil1800_card_wbuffer_inc_pt((_wbuff_p), (_wbuff_p)->head);   \
}while(0)

/* Function called by a workqueue: write data from kernel buffer to
   hardware fifo. This is used if "pause" command is disabled */
extern void galil1800_card_wbuffer_flush(void *);

/* Function called by a workqueue: write data from kernel buffer to
   hardware fifo. This is used if "pause" command is enabled */
extern void galil1800_card_wbuffer_flush_whit_pause_cmd(void *);

/* Function called by "pause" timer to wake up the workqueue */
static inline void galil1800_card_pause_wake_up(unsigned long data)
{
	struct galil1800_card_wbuffer *wbuff_p =
			(struct galil1800_card_wbuffer *)data;
	wbuff_p->pause = 0;
	wake_up(&wbuff_p->read_wait);
}

#define galil1800_add_pause_timer(_wbuff_p) do {                            \
	if(time_before(jiffies,(_wbuff_p)->pause))                          \
		mod_timer(&((_wbuff_p)->pause_timer), (_wbuff_p)->pause);   \
	else                                                                \
		(_wbuff_p)->pause = 0UL;                                    \
}while(0)

/* Check parameter for card device and eventually correct it
   Return 0 if no correction is needed, positive value othewise */
extern int galil1800_card_parameter_check(unsigned int*, unsigned int*);

/* Add card cdev device and register it in sysfs */
extern int galil1800_card_cdev_add(struct cdev *, unsigned int, unsigned int);

/* Delete card cdev device and deregister it in sysfs */
extern void galil1800_card_cdev_del(struct cdev *);

#endif
