/* Copyright (C) 2005-2006  Emiliano Betti and Marco Cesati
   Copyright (C) 2003-2004  Marco Cesati
   Copyright (C) 2002-2003  Shell Technologies s.r.l.

   Authors:
            Emiliano Betti <emilianobetti@gmail.com>
            Marco Cesati   <marcocesati@gmail.com>

   Driver for the Galil 1800 PCI motion controllers (http://www.galilmc.com/)

   This is a driver for the Linux kernel 2.6. It is based on an older driver
   for the kernel 2.4, which in turn was loosely inspired by Galil's own
   Linux driver, as well as by a bunch of PCI drivers in the kernel source code

   This program is free software; you can redistribute  it and/or modify it
   under  the terms of  the GNU General  Public License as published by the
   Free Software Foundation.
*/

#include "galil1800_core.h"
#include "galil1800_io.h"


/* Get serial number
   Return < 0 in case of errors, 0 otherwise */
static int __init galil1800_get_serial_number(struct galil1800_state *s)
{
	const char *lf0 = "LF0\r";
	const char *vf80 = "VF 8.0\r"; /* Conform to GALIL1800_SERIAL_MAXLEN */
	const char *mg_bn = "MG_BN\r"; /* See command 'BN' in the Galil's
					  command reference manual */
	const char *lf1 = "LF1\r";
	const char *vf104 = "VF 10.4\r"; /* Default */

	char buf[GALIL1800_SERIAL_MAXLEN + 10], *p = s->serial;
	int n, i;

	galil1800_send(s, lf0, strlen(lf0));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 10);
	if (!n || buf[n-1] != ':')
		return -EIO;

	galil1800_send(s, vf80, strlen(vf80));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 10);
	if (!n || buf[n-1] != ':')
		return -EIO;

	galil1800_send(s, mg_bn, strlen(mg_bn));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 10);
	if (!n || buf[n-1] != ':' || n-3 > GALIL1800_SERIAL_MAXLEN)
		return -EIO;

	for (i=0; i<n-3; ++i)
		*p++ = buf[i];
	*p = '\0';

	galil1800_send(s, lf1, strlen(lf1));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 10);
	if (!n || buf[n-1] != ':')
		return -EIO;

	galil1800_send(s, vf104, strlen(vf104));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 10);
	if (!n || buf[n-1] != ':')
		return -EIO;

	return 0;
}


/* Print Cars model, revision and firmware
   Return < 0 if errors, 0 otherwise */
static int __init galil1800_get_info(struct galil1800_state *s)
{
	const char *RV = "\x12\x16\r";
	const char *CW = "CW0\r";
	char buf[128];
	int n;

	galil1800_send(s, RV, strlen(RV));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 128);
	if(n>3) {
		buf[n-3] = '\0';
		printk(KERN_INFO PFX "Card model / revision: %s\n", buf);
	}

	galil1800_send(s, CW, strlen(CW));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 128);
	if(n>3) {
		buf[n-3]='\0';
		for (n=0; n<strlen(buf); ++n)
			if (buf[n] == '\n' || buf[n] == '\r')
				buf[n] = ' ';
		printk(KERN_INFO PFX "Firmware: %s\n", buf);
	}

	return galil1800_get_serial_number(s);
}


/* Initialize the controller and some fields of "struct galil1800_state"
   Used in PCI probe function
   Return < 0 if errors, 0 otherwise */
int __init galil1800_init_controller_probe_step(struct galil1800_state *s)
{
	const char *RS = "RS\r";
	/* const char *MASTER_RS = "\x12\x13\r"; */
	const char *EO0 = "EO0\r";
	const char *EO1 = "EO1\r";
	unsigned long j;
	char buf[GALIL1800_READ_FIFO_SIZE+1];
	int ret, n;

	tasklet_disable(&s->irq_tasklet);

	galil1800_reset_fifo_rw(s);

	/* check if FIFO is really flushed */
	n = galil1800_may_write(s);
	if (0==n || 1==n) {
		ret = -EIO;
		goto err;
	}

	galil1800_send(s, RS, strlen(RS));
	j = jiffies + 5*HZ; /* 5 second timeout */

	n=0;

	/* RS can return ":" or "RS\r:" if echo is enabled */
	do {
		ret=galil1800_recv(s,buf+n,GALIL1800_READ_FIFO_SIZE-n);
		if(!ret) {
			schedule();
			continue;
		}
		n+=ret;
		if (buf[n-1] == ':')
			break;
	}while(time_before(jiffies,j));

	if (!n) {
		printk(KERN_NOTICE PFX "RS command timed out."
				" Performing hardware reset\n");
		galil1800_reset_hardware(s);
	} else if (buf[n-1] != ':') {
		buf[n]='\0';
		printk(KERN_ERR PFX "RS command failed! It returned \"%s\"\n",
									buf);
		ret = -EIO;
		goto err;
	}

	/* Disable echo */
	galil1800_send(s, EO0, strlen(EO0));
	schedule_timeout_uninterruptible(HZ/10);
	n = galil1800_recv(s, buf, 10);
	if (!n || buf[n-1] != ':')
		return -EIO;

	ret = galil1800_get_info(s);
	if(ret)
		goto err;

	if(buf[0] == 'E') { /* Echo was enabled */
		GALIL_PRINT_DEBUG("Echo enabled in Flash EEPROM memory\n");
		/* Enable echo */
		galil1800_send(s, EO1, strlen(EO1));
		schedule_timeout_uninterruptible(HZ/10);
		n = galil1800_recv(s, buf, 10);
		if (!n || buf[n-1] != ':')
			return -EIO;
	}

	return 0;
err:
	tasklet_kill(&s->irq_tasklet);
	return ret;
}

/* Initialize the controller and some fields of "struct galil1800_state"
   Only the first opened device call this function
   Return < 0 if errors, 0 otherwise */
static int __galil1800_init_controller_open_step(struct galil1800_state *s)
{
	int result;

	galil1800_reset_fifo_rw(s);

	result = request_irq(s->irq, galil1800_interrupt, SA_SHIRQ,
						GALIL1800_MODULE_NAME, s);

	if (result == -EINVAL) {
		printk(KERN_ERR PFX "bad irq number or handler\n");
		return result;
	}
	if (result == -EBUSY) {
		printk(KERN_ERR PFX "irq %d busy\n", s->irq);
		return result;
	}
	if (result < 0) {
		printk(KERN_ERR PFX "request_irq failed\n");
		return result;
	}

	init_galil1800_irq_stack(&s->irq_stack);
	tasklet_enable(&s->irq_tasklet);

	galil1800_enable_IRQ(s);

	GALIL_PRINT_DEBUG_PID("init controller success!\n");

	return 0;
}

/* Increment open device counter and if it was 0
   initialize the controller and some fields of "struct galil1800_state"
   All open device function must call this function!
   Return < 0 if errors, 0 otherwise */
int galil1800_init_controller_open_step(struct galil1800_state *s)
{
	down(&s->open_devs_sem);

	GALIL_PRINT_DEBUG_PID("other open devs %d\n", s->open_devs);

	if (!s->open_devs) {
		int ret;
		ret = __galil1800_init_controller_open_step(s);
		if(ret < 0) {
			up(&s->open_devs_sem);
			printk(KERN_ERR PFX "init controller failed!\n");
			return ret;
		}
	}
	s->open_devs++;
	up(&s->open_devs_sem);
	return 0;
}

/* Only the last released device call this function */
static void __galil1800_clean_controller_open_step(struct galil1800_state *s)
{
	galil1800_disable_IRQ(s);
	synchronize_irq(s->irq);
	free_irq(s->irq, s);
	tasklet_disable(&s->irq_tasklet);
	GALIL_PRINT_DEBUG_PID("Clean controller performed\n");
}

/* All release device function must call this function! */
void galil1800_clean_controller_open_step(struct galil1800_state *s)
{
	down(&s->open_devs_sem);

	if (--s->open_devs == 0)
		__galil1800_clean_controller_open_step(s);

	GALIL_PRINT_DEBUG_PID("remaining open devs %d\n", s->open_devs);

	up(&s->open_devs_sem);
}
